#![allow(clippy::type_complexity)]
use crate::{levels::stored::*, AppState};

use bevy::{
	input::{keyboard::KeyCode, Input},
	prelude::{
		shape::{Circle, Quad},
		*,
	},
	sprite::Mesh2dHandle,
};
use bevy_mod_picking::*;

pub struct EditorPlugin;

impl Plugin for EditorPlugin {
	fn build(&self, app: &mut App) {
		app.add_event::<DragEndEvent>()
			.add_plugins(DefaultPickingPlugins)
			.add_system_set(SystemSet::on_enter(AppState::Editor).with_system(setup))
			.add_system_set(
				SystemSet::on_update(AppState::Editor)
					.with_system(keyboard_input_system)
					.with_system(follow_ends_system),
			);
	}
}

// Events

struct DragEndEvent(Entity);

// Components

#[derive(Component)]
struct Platform(Entity, Entity);

#[derive(Component)]
struct Draggable;

#[derive(Component)]
struct End(Entity);

// Bundles

#[derive(Bundle)]
struct PlatformBundle {
	#[bundle]
	mesh: ColorMesh2dBundle,
	//platform: Platform,
}

#[derive(Bundle)]
struct PlatformEndBundle {
	#[bundle]
	mesh: ColorMesh2dBundle,
	#[bundle]
	pickable: PickableBundle,
	draggable: Draggable,
	end: End,
}

// Functions

fn spawn_platform(
	commands: &mut Commands,
	meshes: &mut ResMut<Assets<Mesh>>,
	materials: &mut ResMut<Assets<ColorMaterial>>,

	transform: Transform,
	size: Vec2,
) {
	let platform = commands
		.spawn_bundle(PlatformBundle {
			mesh: ColorMesh2dBundle {
				mesh: meshes.add(Mesh::from(Quad { size, flip: false })).into(),
				material: materials.add(ColorMaterial::from(Color::GRAY)),
				transform,
				..default()
			},
		})
		.id();
	let ends = Platform(
		commands
			.spawn_bundle(PlatformEndBundle {
				mesh: ColorMesh2dBundle {
					mesh: meshes
						.add(Mesh::from(Circle {
							radius: 8.,
							vertices: 12,
						}))
						.into(),
					material: materials.add(ColorMaterial::from(Color::rgba(1., 1., 0., 0.7))),
					transform: Transform::from_xyz(
						transform.translation.x - size.x / 2.,
						transform.translation.y - size.y / 2.,
						0.5,
					),
					..default()
				},
				pickable: PickableBundle::default(),
				draggable: Draggable,
				end: End(platform),
			})
			.id(),
		commands
			.spawn_bundle(PlatformEndBundle {
				mesh: ColorMesh2dBundle {
					mesh: meshes
						.add(Mesh::from(Circle {
							radius: 8.,
							vertices: 12,
						}))
						.into(),
					material: materials.add(ColorMaterial::from(Color::rgba(1., 1., 0., 0.7))),
					transform: Transform::from_xyz(
						transform.translation.x + size.x / 2.,
						transform.translation.y + size.y / 2.,
						0.5,
					),
					..default()
				},
				pickable: PickableBundle::default(),
				draggable: Draggable,
				end: End(platform),
			})
			.id(),
	);
	commands.entity(platform).insert(ends);
}

pub fn spawn_stored_level(
	commands: &mut Commands,
	meshes: &mut ResMut<Assets<Mesh>>,
	materials: &mut ResMut<Assets<ColorMaterial>>,
	asset_server: &Res<AssetServer>,

	stored_level: &StoredLevel,
) {
	let _font = asset_server.get_handle::<Font, _>("UacariLegacy-Thin.ttf");

	for platform in stored_level.platforms.iter() {
		spawn_platform(
			commands,
			meshes,
			materials,
			Transform::from_xyz(platform.pos.x, platform.pos.y, 0.),
			platform.size,
		);
	}
}

// Systems

fn setup(
	mut commands: Commands,
	mut meshes: ResMut<Assets<Mesh>>,
	mut materials: ResMut<Assets<ColorMaterial>>,
	camera_query: Query<Entity, With<Camera>>,
	level_id: Res<crate::game::FirstLevel>,
	stored_levels_assets: Res<Assets<StoredLevels>>,
	stored_levels_handle: Res<Handle<StoredLevels>>,
	asset_server: Res<AssetServer>,
) {
	commands
		.entity(camera_query.single())
		.insert_bundle(PickingCameraBundle::default());

	if let Some(stored_level) = stored_levels_assets
		.get(&stored_levels_handle)
		.unwrap()
		.levels
		.get(level_id.0 .0 as usize)
	{
		spawn_stored_level(
			&mut commands,
			&mut meshes,
			&mut materials,
			&asset_server,
			stored_level,
		);
	}
}

fn keyboard_input_system(
	keyboard_input: Res<Input<KeyCode>>,
	mut drag_query: Query<(&mut Transform, &Selection, Option<&End>), With<Draggable>>,
	mut drag_end_event: EventWriter<DragEndEvent>,
) {
	let drag = if keyboard_input.pressed(KeyCode::LShift) || keyboard_input.pressed(KeyCode::RShift)
	{
		Vec3 {
			x: (keyboard_input.pressed(KeyCode::Right) as i8
				- keyboard_input.pressed(KeyCode::Left) as i8) as _,
			y: (keyboard_input.pressed(KeyCode::Up) as i8
				- keyboard_input.pressed(KeyCode::Down) as i8) as _,
			z: 0.,
		}
	} else {
		Vec3 {
			x: (keyboard_input.just_pressed(KeyCode::Right) as i8
				- keyboard_input.just_pressed(KeyCode::Left) as i8) as _,
			y: (keyboard_input.just_pressed(KeyCode::Up) as i8
				- keyboard_input.just_pressed(KeyCode::Down) as i8) as _,
			z: 0.,
		}
	} * 8.;
	if drag != Vec3::ZERO {
		for (mut transform, selection, end) in drag_query.iter_mut() {
			if selection.selected() {
				transform.translation += drag;
				if let Some(End(entity)) = end {
					drag_end_event.send(DragEndEvent(*entity));
				}
			}
		}
	}
}

fn follow_ends_system(
	mut meshes: ResMut<Assets<Mesh>>,
	mut platform_query: Query<(&mut Transform, &mut Mesh2dHandle, &Platform)>,
	end_query: Query<&Transform, Without<Platform>>,
	mut drag_end_event: EventReader<DragEndEvent>,
) {
	for DragEndEvent(entity) in drag_end_event.iter() {
		if let Ok((mut transform, mut mesh, Platform(end1, end2))) = platform_query.get_mut(*entity)
		{
			if let (Ok(end1), Ok(end2)) = (end_query.get(*end1), end_query.get(*end2)) {
				transform.translation.x = (end1.translation.x + end2.translation.x) / 2.;
				transform.translation.y = (end1.translation.y + end2.translation.y) / 2.;
				*mesh = meshes
					.add(Mesh::from(Quad {
						size: Vec2 {
							x: (end2.translation.x - end1.translation.x).abs(),
							y: (end2.translation.y - end1.translation.y).abs(),
						},
						flip: false,
					}))
					.into();
			}
		}
	}
}
